import os.path
from operator import attrgetter

from docx import Document

from docxcompose.composer import Composer
from docxcompose.utils import xpath


XPATH_CACHED_DOCPROPERTY_VALUES = (
    'w:r[preceding-sibling::w:r/w:fldChar/@w:fldCharType="separate"]/w:t'
)


def docx_path(filename):
    return os.path.join(os.path.dirname(__file__), "docs", filename)


def simple_field_expression(name):
    return ".//w:fldSimple[contains(@w:instr, 'DOCPROPERTY \"{}\"')]//w:t".format(name)


def complex_field_expression(name):
    return ".//w:instrText[contains(.,'DOCPROPERTY \"{}\"')]".format(name)


def cached_complex_field_values(element):
    value_elements = xpath(element, XPATH_CACHED_DOCPROPERTY_VALUES)
    return [each.text for each in value_elements]


def assert_simple_field_value(expected, element, name):
    prop_elements = xpath(element, simple_field_expression(name))
    assert len(prop_elements) == 1, 'Could not find simple field "{}"'.format(name)
    actual = prop_elements[0].text
    assert expected == actual, "{} == {}".format(expected, actual)


def assert_complex_field_value(expected, element, name):
    prop_elements = xpath(element, complex_field_expression(name))
    assert len(prop_elements) == 1, 'Could not find complex field "{}"'.format(name)
    parent_paragraph = prop_elements[0].getparent().getparent()
    actual = "".join(cached_complex_field_values(parent_paragraph))
    assert expected == actual, "{} == {}".format(expected, actual)


class ComparableDocument(object):
    """Test helper to compare two docx documents."""

    def __init__(self, doc):
        self.has_neq_partnames = False
        self.neq_parts = []

        self.doc = doc
        if not doc:
            self.parts = []
            self.partnames = []
            return

        self.parts = sorted(self.doc.part.package.parts, key=attrgetter("partname"))
        self.partnames = sorted(p.partname for p in self.parts)

    def __eq__(self, other):
        self.has_neq_partnames = self.partnames != other.partnames
        other.has_neq_partnames = self.has_neq_partnames
        if self.has_neq_partnames:
            return False

        for my_part, other_part in zip(self.parts, other.parts):
            if my_part.blob != other_part.blob:
                self.neq_parts.append((my_part, other_part))
                other.neq_parts.append((other_part, my_part))
        if self.neq_parts:
            return False

        return True

    def post_compare_failed(self, other):
        """Called after a failed comparison/assert."""

        pass


class FixtureDocument(ComparableDocument):
    """Load a comparable document from the composed assets."""

    def __init__(self, composed_filename):
        self.composed_filename = composed_filename

        path = docx_path(os.path.join("composed_fixture", composed_filename))
        doc = Document(path) if os.path.isfile(path) else None

        super(FixtureDocument, self).__init__(doc)


class ComposedDocument(ComparableDocument):
    """Compose at least two documents and provide a docx document for
    comparison.

    Store output document in the `composed_debug` folder when compared to a
    document from the fixture and the assertion failed.

    """

    def __init__(self, master_filename, *filenames):
        composer = Composer(Document(docx_path(master_filename)))
        for filename in filenames:
            composer.append(Document(docx_path(filename)))

        super(ComposedDocument, self).__init__(composer.doc)

    def post_compare_failed(self, other):
        """When comparison to a document from the fixture failed store our
        result in a debug folder.

        """
        if isinstance(other, FixtureDocument):
            path = docx_path(os.path.join("composed_debug", other.composed_filename))
            self.doc.save(path)



from docx import Document
from utils import ComparableDocument
from utils import docx_path
from utils import FixtureDocument

from docxcompose.sdt import StructuredDocumentTags


def test_set_sdt_text_content():
    doc = Document(docx_path("content_controls.docx"))
    sdt = StructuredDocumentTags(doc)
    sdt.set_text("cc.plain_text", "Foo Bar")
    sdt.set_text("cc.plain_text_multiline", "Foo Bar")
    sdt.set_text("cc.plain_text_empty", "Foo Bar")
    sdt.set_text("cc.rich_text", "Foo Bar")

    updated = ComparableDocument(doc)
    expected = FixtureDocument("content_controls.docx")

    assert updated == expected


def test_set_sdt_multiline_text_content():
    doc = Document(docx_path("content_controls_multiline_formatted.docx"))
    sdt = StructuredDocumentTags(doc)
    sdt.set_text("cc.plain_text", "Line 1\nLine 2")
    sdt.set_text("cc.plain_text_multiline", "Line 1\nLine 2")
    sdt.set_text("cc.plain_text_empty", "Line 1\nLine 2")
    sdt.set_text("cc.rich_text", "Line 1\nLine 2")

    updated = ComparableDocument(doc)
    expected = FixtureDocument("content_controls_multiline_formatted.docx")

    assert updated == expected


def test_get_sdt_multiline_text_content():
    doc = FixtureDocument("content_controls_multiline_formatted.docx")
    sdt = StructuredDocumentTags(doc.doc)

    assert sdt.get_text("cc.plain_text") == "Line 1 Line 2"
    assert sdt.get_text("cc.plain_text_multiline") == "Line 1\nLine 2"
    assert sdt.get_text("cc.plain_text_empty") == "Line 1 Line 2"



from datetime import datetime
from datetime import timezone

import docx
import pytest
from docx import Document
from docx.opc.constants import RELATIONSHIP_TYPE as RT
from docx.oxml import parse_xml
from lxml.etree import tostring
from utils import assert_complex_field_value
from utils import assert_simple_field_value
from utils import cached_complex_field_values
from utils import docx_path
from utils import simple_field_expression

from docxcompose.properties import ComplexField
from docxcompose.properties import CUSTOM_PROPERTY_TYPES
from docxcompose.properties import CustomProperties
from docxcompose.properties import SimpleField
from docxcompose.properties import value2vt
from docxcompose.properties import vt2value
from docxcompose.utils import xpath


class TestIdentifyDocpropertiesInDocument(object):
    def test_identifies_simple_fields_correctly(self):
        document = Document(docx_path("outdated_docproperty_with_umlauts.docx"))
        properties = CustomProperties(document).find_docprops_in_document()

        assert 1 == len(properties), "input should contain 1 simple field docproperty"

        prop = properties[0]
        assert prop.name == "F\xfc\xfc"
        assert isinstance(prop, SimpleField)

    def test_identifies_complex_fields_correctly(self):
        document = Document(docx_path("three_props_in_same_paragraph.docx"))
        properties = CustomProperties(document).find_docprops_in_document()

        assert len(document.paragraphs) == 1, "input file should contains one paragraph"
        assert len(properties) == 3, (
            "input should contain three complex field docproperties"
        )

        # check that all fields were identified as complex fields
        for prop in properties:
            assert isinstance(prop, ComplexField)

        # check that all field names were parsed correctly
        expected_names = ("Text Property", "Number Property", "Text Property")
        for name, prop in zip(expected_names, properties):
            assert name == prop.name

        # check that begin, separate and end were identified correctly
        attrib_key = (
            "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}fldCharType"
        )
        expected_indexes = ((1, 3, 11), (13, 15, 17), (25, 27, 32))
        for prop, indexes in zip(properties, expected_indexes):
            assert prop.begin_run.getchildren()[1].attrib[attrib_key] == "begin"
            assert (
                prop.get_separate_run().getchildren()[1].attrib[attrib_key]
                == "separate"
            )
            assert prop.end_run.getchildren()[1].attrib[attrib_key] == "end"
            assert prop.w_p.index(prop.begin_run) == indexes[0]
            assert prop.w_p.index(prop.get_separate_run()) == indexes[1]
            assert prop.w_p.index(prop.end_run) == indexes[2]

    def test_finds_run_nodes_in_complex_fields_correctly(self):
        document = Document(docx_path("three_props_in_same_paragraph.docx"))
        properties = CustomProperties(document).find_docprops_in_document()

        assert len(properties) == 3, (
            "input should contain three complex field docproperties"
        )

        # In the first field, there are the following runs: begin, docprop,
        # a separate, 3 runs for the value (because of spellcheck) and end
        prop = properties[0]
        assert len(prop.get_runs_for_update()) == 3

        # when dissolving a property the runs that have to be removed
        # are the begin, docprop, and from separate to end (here just
        # separate and end)
        runs = prop.get_runs_to_replace_field_with_value()
        assert len(runs) == 4
        assert runs[0] == prop.begin_run
        assert runs[1] == prop.w_r
        assert runs[2] == prop.get_separate_run()
        assert runs[-1] == prop.end_run

        # In the second field, there are the following runs: begin, docprop,
        # a separate, 1 run for the value and end
        prop = properties[1]
        assert len(prop.get_runs_for_update()) == 1

        # when dissolving a property the runs that have to be removed
        # are from begin to separate (here begin, docprop, separate)
        # and the end
        runs = prop.get_runs_to_replace_field_with_value()
        assert len(runs) == 4
        assert runs[0] == prop.begin_run
        assert runs[1] == prop.w_r
        assert runs[2] == prop.get_separate_run()
        assert runs[-1] == prop.end_run

        # In the first field, there are the following runs: begin, docprop,
        # a separate, 2 runs for the value (because of spellcheck) and end
        prop = properties[2]
        assert len(prop.get_runs_for_update()) == 2

        # when dissolving a property the runs that have to be removed
        # are the begin, docprop, and from separate to end (here just
        # separate and end)
        runs = prop.get_runs_to_replace_field_with_value()
        assert len(runs) == 4
        assert runs[0] == prop.begin_run
        assert runs[1] == prop.w_r
        assert runs[2] == prop.get_separate_run()
        assert runs[-1] == prop.end_run

    def test_finds_run_nodes_in_complex_field_without_separate_correctly(self):
        document = Document(docx_path("complex_field_without_separate.docx"))
        properties = CustomProperties(document).find_docprops_in_document()

        assert len(properties) == 2, (
            "input should contain two complex field docproperties"
        )

        # The "User.FullName" docproperty should be the one without a separate run
        # In this field, there are the following runs: begin, docprop and end
        matches = [prop for prop in properties if prop.name == "User.FullName"]
        assert len(matches) == 1, "There should be only one User.FullName docproperty"
        prop = matches[0]
        assert prop.get_separate_run() is None, (
            "This complex field should not have a separate run."
        )
        assert prop.get_runs_for_update() == [], (
            "As there is no separate run, there should be no run to update"
        )

        # As there are no separate, all runs should be removed when dissolving
        # the property.
        runs = prop.get_runs_to_replace_field_with_value()
        assert len(runs) == 3
        assert runs[0] == prop.begin_run
        assert runs[1] == prop.w_r
        assert runs[2] == prop.end_run

    def test_finds_field_name_and_format_when_split_in_several_runs(self):
        document = Document(docx_path("complex_field_with_split_fieldname.docx"))
        properties = CustomProperties(document).find_docprops_in_document()

        assert len(properties) == 1, (
            "input should contain one complex field docproperties"
        )

        prop = properties[0]

        # make sure that the field name and format are indeed split over several runs
        assert [
            [each.text for each in xpath(run, "w:instrText")] for run in prop._runs[:9]
        ] == [
            [' DOCPROPERTY "ogg.document.document_date"  \\@ "ddd'],
            ["d"],
            [" dd"],
            [" "],
            ["MM"],
            ["M"],
            ["M"],
            [" yy"],
            ['yy hh:mm:s" \\* MERGEFORMAT '],
        ]

        assert (
            prop._get_fieldname_string()
            == ' DOCPROPERTY "ogg.document.document_date"  \\@ "ddd'
            + 'd dd MMMM yyyy hh:mm:s" \\* MERGEFORMAT '
        )

        assert prop.name == "ogg.document.document_date"
        assert prop.date_format == "EEEE dd MMMM yyyy hh:mm:s"

    def test_finds_header_footer_of_different_sections(self):
        document = Document(docx_path("docproperties_header_footer_3_sections.docx"))
        properties = CustomProperties(document).find_docprops_in_document()

        assert len(properties) == 5, (
            "input should contain 5 properties in header/footer and body"
        )

        expected_properties = [
            "Text Property",
            "Number Property",
            "Date Property",
            "Float Property",
            "Boolean Property",
        ]

        assert [each.name for each in properties] == expected_properties


class TestUpdateAllDocproperties(object):
    def test_updates_doc_properties_in_header(self):
        document = Document(docx_path("docproperties_header.docx"))

        assert_simple_field_value(
            "xxx", document.sections[0].header.part.element, "my.text-prop"
        )

        CustomProperties(document).update_all()

        assert_simple_field_value(
            "i'm some text", document.sections[0].header.part.element, "my.text-prop"
        )

    def test_updates_doc_properties_in_footer(self):
        document = Document(docx_path("docproperties_footer.docx"))

        assert_simple_field_value(
            "xxx", document.sections[0].footer.part.element, "my.text-prop"
        )

        CustomProperties(document).update_all()

        assert_simple_field_value(
            "b\xe4hh", document.sections[0].footer.part.element, "my.text-prop"
        )

    def test_updates_doc_properties_different_first_page(self):
        document = Document(
            docx_path("docproperties_different_first_page_1_section.docx")
        )

        assert_simple_field_value(
            "xxx", document.sections[0].first_page_header.part.element, "page1.header"
        )
        assert_simple_field_value(
            "xxx", document.sections[0].first_page_footer.part.element, "page1.footer"
        )
        assert_simple_field_value(
            "0", document.sections[0].header.part.element, "page2.header"
        )
        assert_simple_field_value(
            "01.01.1970", document.sections[0].footer.part.element, "page2.footer"
        )

        CustomProperties(document).update_all()

        assert_simple_field_value(
            "p1h", document.sections[0].first_page_header.part.element, "page1.header"
        )
        assert_simple_field_value(
            "p1f", document.sections[0].first_page_footer.part.element, "page1.footer"
        )
        assert_simple_field_value(
            "42", document.sections[0].header.part.element, "page2.header"
        )
        assert_simple_field_value(
            "18.10.1984", document.sections[0].footer.part.element, "page2.footer"
        )

    def test_updates_doc_properties_different_odd_even_pages(self):
        document = Document(
            docx_path("docproperties_different_odd_even_pages_1_section.docx")
        )

        assert_simple_field_value(
            "xxx", document.sections[0].header.part.element, "odd.header"
        )
        assert_simple_field_value(
            "xxx", document.sections[0].footer.part.element, "odd.footer"
        )
        assert_simple_field_value(
            "0", document.sections[0].even_page_header.part.element, "even.header"
        )
        assert_simple_field_value(
            "Y", document.sections[0].even_page_footer.part.element, "even.footer"
        )

        CustomProperties(document).update_all()

        assert_simple_field_value(
            "odd-header", document.sections[0].header.part.element, "odd.header"
        )
        assert_simple_field_value(
            "odd-footer", document.sections[0].footer.part.element, "odd.footer"
        )
        assert_simple_field_value(
            "1337", document.sections[0].even_page_header.part.element, "even.header"
        )
        assert_simple_field_value(
            "N", document.sections[0].even_page_footer.part.element, "even.footer"
        )

    def test_updates_docproperties_shared_footer(self):
        document = Document(docx_path("docproperties_shared_footer_2_sections.docx"))

        header_1 = document.sections[0].header
        header_2 = document.sections[1].header
        assert not header_1.is_linked_to_previous
        assert not header_2.is_linked_to_previous
        assert_complex_field_value("xxx", header_1.part.element, "section1.header")
        assert_complex_field_value("yyy", header_2.part.element, "section2.header")

        # the same footer should be referenced by both sections
        # the sections should be considered as linked
        footer_1 = document.sections[0].footer
        footer_2 = document.sections[1].footer
        assert not footer_1.is_linked_to_previous
        assert footer_2.is_linked_to_previous
        assert_complex_field_value("99", footer_1.part.element, "shared.footer")
        assert_complex_field_value("99", footer_2.part.element, "shared.footer")

        CustomProperties(document).update_all()

        assert_complex_field_value(
            "h\xe4der 1", header_1.part.element, "section1.header"
        )
        assert_complex_field_value(
            "h\xf6der 2", header_2.part.element, "section2.header"
        )

        # the same footer should be referenced by both sections
        footer_1 = document.sections[0].footer
        footer_2 = document.sections[1].footer
        assert_complex_field_value("123123123", footer_1.part.element, "shared.footer")
        assert_complex_field_value("123123123", footer_2.part.element, "shared.footer")

    def test_updates_docproperties_shared_header(self):
        document = Document(docx_path("docproperties_shared_header_2_sections.docx"))

        # the same header should be referenced by both sections
        # the sections should be considered as linked
        header_1 = document.sections[0].header
        header_2 = document.sections[1].header
        assert not header_1.is_linked_to_previous
        assert header_2.is_linked_to_previous
        assert_complex_field_value("xxx", header_1.part.element, "shared.header")
        assert_complex_field_value("xxx", header_2.part.element, "shared.header")

        footer_1 = document.sections[0].footer
        footer_2 = document.sections[1].footer
        assert not footer_1.is_linked_to_previous
        assert not footer_2.is_linked_to_previous
        assert_complex_field_value("123", footer_1.part.element, "section1.footer")
        assert_complex_field_value("yyy", footer_2.part.element, "section2.footer")

        CustomProperties(document).update_all()

        # the same header should be referenced by both sections
        assert_complex_field_value("sh\xe4red", header_1.part.element, "shared.header")
        assert_complex_field_value("sh\xe4red", header_2.part.element, "shared.header")

        footer_1 = document.sections[0].footer
        footer_2 = document.sections[1].footer
        assert_complex_field_value("-1.0", footer_1.part.element, "section1.footer")
        assert_complex_field_value(
            "f\xfc\xfcter", footer_2.part.element, "section2.footer"
        )

    def test_updates_docproperties_shared_header_footer(self):
        document = Document(
            docx_path("docproperties_shared_header_footer_2_sections.docx")
        )

        # the same header should be referenced by both sections
        # the sections should be considered as linked
        header_1 = document.sections[0].header
        header_2 = document.sections[1].header
        assert not header_1.is_linked_to_previous
        assert header_2.is_linked_to_previous
        assert_complex_field_value("blub", header_1.part.element, "shared.header")
        assert_complex_field_value("blub", header_2.part.element, "shared.header")

        # the same footer should be referenced by both sections
        # the sections should be considered as linked
        footer_1 = document.sections[0].footer
        footer_2 = document.sections[1].footer
        assert not footer_1.is_linked_to_previous
        assert footer_2.is_linked_to_previous
        assert_complex_field_value("N", footer_1.part.element, "shared.footer")
        assert_complex_field_value("N", footer_2.part.element, "shared.footer")

        CustomProperties(document).update_all()

        # the same header should be referenced by both sections
        assert_complex_field_value("ig bi obe", header_1.part.element, "shared.header")
        assert_complex_field_value("ig bi obe", header_2.part.element, "shared.header")

        # the same footer should be referenced by both sections
        footer_1 = document.sections[0].footer
        footer_2 = document.sections[1].footer
        assert_complex_field_value("Y", footer_1.part.element, "shared.footer")
        assert_complex_field_value("Y", footer_2.part.element, "shared.footer")

    def test_updates_doc_properties_with_umlauts(self):
        document = Document(docx_path("outdated_docproperty_with_umlauts.docx"))

        assert_simple_field_value("xxx", document.element.body, "F\xfc\xfc")

        CustomProperties(document).update_all()

        assert_simple_field_value("j\xe4ja.", document.element.body, "F\xfc\xfc")

    def test_complex_docprop_fields_with_multiple_textnodes_are_updated(self):
        document = Document(docx_path("spellchecked_docproperty.docx"))
        paragraphs = xpath(document.element.body, "//w:p")
        assert len(paragraphs) == 1, "input file contains one paragraph"
        assert len(xpath(document.element.body, "//w:instrText")) == 1, (
            "input contains one complex field docproperty"
        )
        w_p = paragraphs[0]

        cached_values = cached_complex_field_values(w_p)
        assert len(cached_values) == 4, "doc property value is scattered over 4 parts"
        assert "".join(cached_values) == "i will be spllchecked!"

        CustomProperties(document).update_all()

        w_p = xpath(document.element.body, "//w:p")[0]
        cached_values = cached_complex_field_values(w_p)
        assert len(cached_values) == 1, (
            "doc property value has been reset to one cached value"
        )
        assert cached_values[0] == "i will be spllchecked!"

    def test_complex_docprop_with_multiple_textnode_in_same_run_are_updated(self):
        document = Document(docx_path("two_textnodes_in_run_docproperty.docx"))
        paragraphs = xpath(document.element.body, "//w:p")
        assert len(paragraphs) == 1, "input file contains one paragraph"
        assert len(xpath(document.element.body, "//w:instrText")) == 1, (
            "input contains one complex field docproperty"
        )

        w_p = paragraphs[0]
        cached_values = cached_complex_field_values(w_p)
        assert len(cached_values) == 2, "doc property value is scattered over 2 parts"
        assert "".join(cached_values) == "Hello there"

        CustomProperties(document).update_all()

        w_p = xpath(document.element.body, "//w:p")[0]
        cached_values = cached_complex_field_values(w_p)
        assert len(cached_values) == 1, (
            "doc property value has been reset to one cached value"
        )
        assert cached_values[0] == "i will be spllchecked!"

    def test_three_complex_docprop_in_same_paragraph(self):
        document = Document(docx_path("three_props_in_same_paragraph.docx"))
        properties = CustomProperties(document)

        assert len(document.paragraphs) == 1, "input file should contains one paragraph"
        paragraph = document.paragraphs[0]
        assert len(properties.find_docprops_in_document()) == 3, (
            "input should contain three complex field docproperties"
        )

        text = "{text} / {num} mor between the fields {text} and some afte the three fields"
        assert paragraph.text == text.format(text="I was spellcecked", num=0)

        properties.update_all()

        assert paragraph.text == text.format(text="Foo", num=2)

    def test_multiple_identical_docprops_get_updated(self):
        document = Document(docx_path("multiple_identical_properties.docx"))
        assert len(document.paragraphs) == 3, "input file should contain 3 paragraphs"
        for paragraph in document.paragraphs:
            assert len(xpath(paragraph._p, ".//w:instrText")) == 1, (
                "paragraph should contain one complex field docproperties"
            )

            assert paragraph.text == "Foo"

        CustomProperties(document).update_all()

        for i, paragraph in enumerate(document.paragraphs):
            assert paragraph.text == "Bar", "docprop {} was not updated".format(i + 1)

    def test_docproperty_without_separate_does_get_updated(self):
        document = Document(docx_path("complex_field_without_separate.docx"))
        custom_properties = CustomProperties(document)
        properties = custom_properties.find_docprops_in_document()
        paragraphs = document.paragraphs

        # Make sure that a value is set for 'User.FullName'
        assert ("User.FullName", "Test User") in custom_properties.items()
        assert ("Dossier.Title", " Some Title") in custom_properties.items()

        # Make sure that 'User.FullName' field has no separate node
        matches = [prop for prop in properties if prop.name == "User.FullName"]
        assert len(matches) == 1, "There should be only one User.FullName docproperty"
        fullname = matches[0]
        assert fullname.get_separate_run() is None, (
            "This complex field should not have a separate run."
        )

        # Make sure that 'Dossier.Title' field has a separate node
        matches = [prop for prop in properties if prop.name == "Dossier.Title"]
        assert len(matches) == 1, "There should be only one Dossier.Title docproperty"
        title = matches[0]
        assert title.get_separate_run() is not None, (
            "This complex field should have a separate run."
        )

        # Check the content of the paragraphs before update
        assert len(paragraphs) == 2
        assert paragraphs[0].text == "Sachbearbeiter: "
        assert "Sachbearbeiter: " in fullname.w_p.xml
        assert paragraphs[1].text == "Dossier Titel:  "

        custom_properties.update_all()

        # Field with missing separate was not updated
        assert paragraphs[0].text == "Sachbearbeiter: Test User"
        # Next field was updated correctly
        assert paragraphs[1].text == "Dossier Titel:  Some Title"

    def test_date_docprops_with_format_get_updated(self):
        document = Document(docx_path("date_docproperties_with_format.docx"))
        assert len(document.paragraphs) == 3, "input file should contain 3 paragraph"

        expected_values = ["11.06.19", "mardi 11 juin 2019", "11-6-19 0:0:0"]
        for expected, paragraph in zip(expected_values, document.paragraphs):
            assert paragraph.text == expected

        CustomProperties(document).update_all()

        expected_values = ["23.01.20", "jeudi 23 janvier 2020", "23-1-20 10:0:0"]
        for i, (expected, paragraph) in enumerate(
            zip(expected_values, document.paragraphs)
        ):
            assert paragraph.text == expected, (
                "docprop {} was not updated correctly".format(i + 1)
            )

    def test_date_docprops_respect_language(self):
        document = Document(docx_path("date_docproperties_with_format.docx"))
        assert len(document.paragraphs) == 3, "input file should contain 3 paragraph"

        CustomProperties(document).update_all()
        assert document.paragraphs[1].text == "jeudi 23 janvier 2020"

        document.element.xpath(".//w:lang")[0].set(
            docx.oxml.shared.qn("w:val"), "de-CH"
        )
        CustomProperties(document).update_all()
        assert document.paragraphs[1].text == "Donnerstag 23 Januar 2020"

        document.element.xpath(".//w:lang")[0].set(
            docx.oxml.shared.qn("w:val"), "en-US"
        )
        CustomProperties(document).update_all()
        assert document.paragraphs[1].text == "Thursday 23 January 2020"

    def test_docprops_with_split_fieldname_get_updated(self):
        document = Document(docx_path("complex_field_with_split_fieldname.docx"))
        assert len(document.paragraphs) == 1, "input file should contain 1 paragraph"

        paragraph = document.paragraphs[0]
        assert paragraph.text == "Datum: Tuesday 09 February 2021 00:00:00"

        CustomProperties(document).update_all()
        assert paragraph.text == "Datum: Friday 11 March 2022 10:00:0"


class TestUpdateSpecificDocproperty(object):
    def test_simple_field_gets_updated(self):
        document = Document(docx_path("outdated_docproperty_with_umlauts.docx"))
        assert_simple_field_value("xxx", document.element.body, "F\xfc\xfc")

        CustomProperties(document).update("F\xfc\xfc", "new v\xe4lue")

        assert_simple_field_value("new v\xe4lue", document.element.body, "F\xfc\xfc")

    def test_complex_field_gets_updated(self):
        document = Document(docx_path("docproperties.docx"))
        assert len(document.paragraphs) == 6, "input file should contain 6 paragraphs"

        properties = xpath(document.element.body, ".//w:instrText")
        assert len(properties) == 5, (
            "input should contain five complex field docproperties"
        )

        expected_paragraphs = [
            "Custom Doc Properties",
            "Text: Foo Bar",
            "Number: 123",
            "Boolean: Y",
            "Date: 11.06.2019",
            "Float: 1.1",
        ]
        actual_paragraphs = [paragraph.text for paragraph in document.paragraphs]
        assert actual_paragraphs == expected_paragraphs

        CustomProperties(document).update("Number Property", 423)

        expected_paragraphs[2] = "Number: 423"
        actual_paragraphs = [paragraph.text for paragraph in document.paragraphs]
        assert actual_paragraphs == expected_paragraphs

    def test_multiple_identical_docprops_get_updated(self):
        document = Document(docx_path("multiple_identical_properties.docx"))
        assert len(document.paragraphs) == 3, "input file should contain 3 paragraphs"
        for paragraph in document.paragraphs:
            assert len(xpath(paragraph._p, ".//w:instrText")) == 1, (
                "paragraph should contain one complex field docproperties"
            )

            assert paragraph.text == "Foo"

        CustomProperties(document).update("Text Property", "New value")

        for i, paragraph in enumerate(document.paragraphs):
            assert paragraph.text == "New value", "docprop {} was not updated".format(
                i + 1
            )


class TestDissolveField(object):
    def test_removes_simple_field_but_keeps_value(self):
        document = Document(docx_path("outdated_docproperty_with_umlauts.docx"))
        assert len(document.paragraphs) == 1, "input file should contain 1 paragraph"
        fields = xpath(document.element.body, simple_field_expression("F\xfc\xfc"))
        assert len(fields) == 1, "should contain one simple field docproperty"

        assert document.paragraphs[0].text == "Hie chund ds property: "
        assert fields[0].text == "xxx"

        CustomProperties(document).dissolve_fields("F\xfc\xfc")
        fields = xpath(document.element.body, simple_field_expression("F\xfc\xfc"))
        assert len(fields) == 0, "should not contain any docproperties anymore"
        # when simple field is removed, the value is moved one up in the hierarchy
        assert document.paragraphs[0].text == "Hie chund ds property: xxx"

    def test_removes_complex_field_but_keeps_value(self):
        # test fails because field has 2 spaces before docprop name
        document = Document(docx_path("docproperties.docx"))
        assert len(document.paragraphs) == 6, "input file should contain 6 paragraphs"

        properties = xpath(document.element.body, ".//w:instrText")
        assert len(properties) == 5, (
            "input should contain five complex field docproperties"
        )

        expected_paragraphs = [
            "Custom Doc Properties",
            "Text: Foo Bar",
            "Number: 123",
            "Boolean: Y",
            "Date: 11.06.2019",
            "Float: 1.1",
        ]
        actual_paragraphs = [paragraph.text for paragraph in document.paragraphs]
        assert actual_paragraphs == expected_paragraphs

        CustomProperties(document).dissolve_fields("Number Property")

        actual_paragraphs = [paragraph.text for paragraph in document.paragraphs]
        assert actual_paragraphs == expected_paragraphs

        properties = xpath(document.element.body, ".//w:instrText")
        assert len(properties) == 4, "only 4 fields should remain after removal of one"

    def test_dissolves_all_instances_of_given_field(self):
        document = Document(docx_path("multiple_identical_properties.docx"))
        assert len(document.paragraphs) == 3, "input file should contain 3 paragraphs"
        assert len(xpath(document.element.body, ".//w:instrText")) == 3, (
            "document should contain three complex field docproperties"
        )

        for paragraph in document.paragraphs:
            assert paragraph.text == "Foo"

        CustomProperties(document).dissolve_fields("Text Property")

        assert len(document.paragraphs) == 3
        assert len(xpath(document.element.body, ".//w:instrText")) == 0, (
            "document should not contain any complex field anymore"
        )
        for paragraph in document.paragraphs:
            assert paragraph.text == "Foo", "value should have been kept in document"

    def test_dissolving_field_when_three_complex_docprop_in_same_paragraph(self):
        document = Document(docx_path("three_props_in_same_paragraph.docx"))
        assert len(document.paragraphs) == 1, "input file should contains one paragraph"
        paragraph = document.paragraphs[0]
        properties = CustomProperties(document)
        assert len(properties.find_docprops_in_document()) == 3, (
            "input should contain three complex field docproperties"
        )

        text = "{text} / {num} mor between the fields {text} and some afte the three fields"
        assert paragraph.text == text.format(text="I was spellcecked", num=0)

        properties.dissolve_fields("Text Property")

        assert len(document.paragraphs) == 1
        assert len(properties.find_docprops_in_document()) == 1, (
            "document should contain one complex field after removal"
        )
        assert paragraph.text == text.format(text="I was spellcecked", num=0)


def test_get_doc_properties():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    assert props["Text Property"] == "Foo Bar"
    assert props["Number Property"] == 123
    assert props["Boolean Property"] is True
    assert props["Date Property"] == datetime(2019, 6, 11, 10, 0, tzinfo=timezone.utc)

    assert props.get("Text Property") == "Foo Bar"
    assert props.get("Number Property") == 123
    assert props.get("Boolean Property") is True
    assert props.get("Date Property") == datetime(
        2019, 6, 11, 10, 0, tzinfo=timezone.utc
    )


def test_get_doc_property_is_case_insensitive():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    assert props["text property"] == "Foo Bar"
    assert props.get("text property") == "Foo Bar"


def test_add_doc_properties():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    props.add("My Text Property", "foo bar")
    assert props.get("My Text Property") == "foo bar"

    props.add("My Boolean Property", True)
    assert props.get("My Boolean Property") is True

    props.add("My Number Property", 123)
    assert props.get("My Number Property") == 123

    props.add(
        "My Date Property", datetime(2019, 10, 23, 15, 44, 50, tzinfo=timezone.utc)
    )
    assert props.get("My Date Property") == datetime(
        2019, 10, 23, 15, 44, 50, tzinfo=timezone.utc
    )


def test_add_utf8_property():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    props.add("My Text Property", "f\xfc\xfc".encode("utf-8"))
    assert props.get("My Text Property") == "f\xfc\xfc"


def test_set_doc_properties():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    props["Text Property"] = "baz"
    assert props["Text Property"] == "baz"

    props["Boolean Property"] = False
    assert props["Boolean Property"] is False

    props["Number Property"] = 456
    assert props["Number Property"] == 456

    props["Date Property"] = datetime(2019, 10, 20, 12, 0, tzinfo=timezone.utc)
    assert props["Date Property"] == datetime(2019, 10, 20, 12, 0, tzinfo=timezone.utc)


def test_set_doc_property_is_case_insensitive():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    props["text property"] = "baz"
    assert props["Text Property"] == "baz"


def test_delete_doc_properties():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    del props["Text Property"]
    del props["Number Property"]

    assert "Text Property" not in props
    assert "Number Property" not in props

    assert xpath(props._element, ".//cp:property/@pid") == ["2", "3", "4"]


def test_delete_doc_property_is_case_insensitive():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    del props["text property"]

    assert "Text Property" not in props


def test_contains_is_case_insensitive():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    assert "text property" in props


def test_nullify_doc_properties():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    props.nullify("Text Property")
    props.nullify("Number Property")
    props.nullify("Boolean Property")
    props.nullify("Date Property")
    props.nullify("Float Property")

    assert props["Text Property"] == ""
    assert "Number Property" not in props
    assert "Boolean Property" not in props
    assert "Date Property" not in props
    assert "Float Property" not in props


def test_nullify_doc_property_is_case_insensitive():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    props.nullify("text property")
    assert props["Text Property"] == ""


def test_set_doc_property_on_document_without_properties_creates_new_part():
    document = Document(docx_path("master.docx"))
    props = CustomProperties(document)
    props["Text Property"] = "Foo"

    assert props.part is not None
    assert props["Text Property"] == "Foo"

    part = document.part.package.part_related_by(RT.CUSTOM_PROPERTIES)
    assert part is not None


def test_doc_properties_keys():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    assert props.keys() == [
        "Text Property",
        "Number Property",
        "Boolean Property",
        "Date Property",
        "Float Property",
    ]


def test_doc_properties_values():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    assert props.values() == [
        "Foo Bar",
        123,
        True,
        datetime(2019, 6, 11, 10, 0, tzinfo=timezone.utc),
        1.1,
    ]


def test_doc_properties_items():
    document = Document(docx_path("docproperties.docx"))
    props = CustomProperties(document)

    assert props.items() == [
        ("Text Property", "Foo Bar"),
        ("Number Property", 123),
        ("Boolean Property", True),
        ("Date Property", datetime(2019, 6, 11, 10, 0, tzinfo=timezone.utc)),
        ("Float Property", 1.1),
    ]


def test_vt2value_value2vt_roundtrip():
    assert vt2value(value2vt(42)) == 42
    assert vt2value(value2vt(True)) is True
    assert vt2value(value2vt(1.1)) == pytest.approx(1.1)
    dt = datetime(2019, 6, 11, 10, 0, tzinfo=timezone.utc)
    assert vt2value(value2vt(dt)) == dt
    assert vt2value(value2vt("foo")) == "foo"
    assert vt2value(value2vt("")) == ""

    node = parse_xml(CUSTOM_PROPERTY_TYPES["int"])
    node.text = "42"
    assert tostring(value2vt(vt2value(node))) == tostring(node)

    node = parse_xml(CUSTOM_PROPERTY_TYPES["bool"])
    node.text = "true"
    assert tostring(value2vt(vt2value(node))) == tostring(node)

    node = parse_xml(CUSTOM_PROPERTY_TYPES["float"])
    node.text = "1.1"
    assert tostring(value2vt(vt2value(node))) == tostring(node)

    node = parse_xml(CUSTOM_PROPERTY_TYPES["datetime"])
    node.text = "2003-12-31T10:14:55Z"
    assert tostring(value2vt(vt2value(node))) == tostring(node)

    node = parse_xml(CUSTOM_PROPERTY_TYPES["text"])
    node.text = "foo"
    assert tostring(value2vt(vt2value(node))) == tostring(node)

    node = parse_xml(CUSTOM_PROPERTY_TYPES["text"])
    node.text = ""
    assert tostring(value2vt(vt2value(node))) == tostring(node)


def test_vt2value_returns_empty_string_for_missing_text_node():
    node = parse_xml(CUSTOM_PROPERTY_TYPES["text"])
    node.text = None
    assert vt2value(node) == ""



try:
    from aiohttp import web
except ImportError:
    raise SystemExit("Install with server extra to use this command.")
import importlib.metadata
import logging
import os.path
import tempfile

from docx import Document

from docxcompose.composer import Composer
from docxcompose.utils import to_bool


CHUNK_SIZE = 65536
logger = logging.getLogger("docxcompose")
version = importlib.metadata.version("docxcompose")


async def compose(request):

    documents = []

    if not request.content_type == "multipart/form-data":
        logger.info(
            "Bad request. Received content type %s instead of multipart/form-data.",
            request.content_type,
        )
        return web.Response(status=400, text="Multipart request required")

    reader = await request.multipart()

    with tempfile.TemporaryDirectory() as temp_dir:
        while True:
            part = await reader.next()

            if part is None:
                break

            if part.filename is None:
                continue

            documents.append(await save_part_to_file(part, temp_dir))

        if not documents:
            return web.Response(status=400, text="No documents provided")

        composed_filename = os.path.join(temp_dir, "composed.docx")

        try:
            composer = Composer(Document(documents.pop(0)), **compose_options(request))
            for document in documents:
                composer.append(Document(document))
            composer.save(composed_filename)
        except Exception:
            logger.exception("Failed composing documents.")
            return web.Response(status=500, text="Failed composing documents")

        return await stream_file(
            request,
            composed_filename,
            "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        )


def compose_options(request):
    return {
        "preserve_styles": to_bool(request.rel_url.query.get("preserve_styles", "")),
    }


async def save_part_to_file(part, directory):
    filename = os.path.join(directory, f"{part.name}_{part.filename}")
    with open(filename, "wb") as file_:
        while True:
            chunk = await part.read_chunk(CHUNK_SIZE)
            if not chunk:
                break
            file_.write(chunk)
    return filename


async def stream_file(request, filename, content_type):
    response = web.StreamResponse(
        status=200,
        reason="OK",
        headers={
            "Content-Type": content_type,
            "Content-Disposition": f'attachment; filename="{os.path.basename(filename)}"',  # noqa
        },
    )
    await response.prepare(request)

    with open(filename, "rb") as outfile:
        while True:
            data = outfile.read(CHUNK_SIZE)
            if not data:
                break
            await response.write(data)

    await response.write_eof()
    return response


async def healthcheck(request):
    return web.Response(status=200, text="OK")


def create_app():
    app = web.Application()
    app.add_routes([web.post("/", compose)])
    app.add_routes([web.get("/healthcheck", healthcheck)])
    return app


def main():
    print(f"docxcompose {version}")
    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(name)s %(message)s",
        level=logging.INFO,
    )
    web.run_app(create_app())


if __name__ == "__main__":
    main()



import random
from difflib import unified_diff

import pytest
from lxml import etree
from utils import ComparableDocument


@pytest.fixture
def static_reseed():
    """Static random seed fixture to test content that generates random ids."""
    random.seed(42)


def pytest_assertrepr_compare(config, op, left, right):
    if (
        isinstance(left, ComparableDocument)
        and isinstance(right, ComparableDocument)
        and op == "=="
    ):
        left.post_compare_failed(right)
        right.post_compare_failed(left)

        if left.has_neq_partnames:
            extra_right = [
                item for item in right.partnames if item not in left.partnames
            ]
            extra_left = [
                item for item in left.partnames if item not in right.partnames
            ]

            explanation = ["documents contain same parts"]
            if right.doc is None:
                explanation.append("Right document is None")
            if left.doc is None:
                explanation.append("Left document is None")
            if extra_left:
                explanation.append(
                    "Left contains extra parts {}".format(", ".join(extra_left))
                )
            if extra_right:
                explanation.append(
                    "Right contains extra parts {}".format(", ".join(extra_right))
                )
            return explanation

        diffs = []
        for lpart, rpart in left.neq_parts:
            if not lpart.partname.endswith(".xml"):
                diffs.append("Binary parts differ {}".format(lpart.partname))
                diffs.append("")
                continue

            doc = etree.fromstring(lpart.blob)
            left_xml = etree.tounicode(doc, pretty_print=True)
            doc = etree.fromstring(rpart.blob)
            right_xml = etree.tounicode(doc, pretty_print=True)

            diffs.extend(
                unified_diff(
                    left_xml.splitlines(),
                    right_xml.splitlines(),
                    fromfile=lpart.partname,
                    tofile=lpart.partname,
                )
            )
            diffs.append("")

        if diffs:
            filenames = [p[0].partname for p in left.neq_parts]
            diffs.insert(
                0,
                "document parts are equal. Not equal parts: {}".format(
                    ", ".join(filenames)
                ),
            )
            return diffs



from datetime import datetime

from babel.dates import format_datetime

from docxcompose.properties import FieldBase
from docxcompose.utils import word_to_python_date_format


class FieldForTesting(FieldBase):
    def _get_fieldname_string(self):
        return self.node


class TestFieldNameParsing(object):
    def test_can_parse_quoted_property_names(self):
        node = ' DOCPROPERTY "Propertyname"  \\* MERGEFORMAT '
        assert FieldForTesting(node).name == "Propertyname"

    def test_can_parse_unquoted_property_names(self):
        node = " DOCPROPERTY Propertyname  \\* MERGEFORMAT "
        assert FieldForTesting(node).name == "Propertyname"

    def test_can_parse_quoted_property_names_with_spaces(self):
        node = ' DOCPROPERTY "Text Property"  \\* MERGEFORMAT '
        assert FieldForTesting(node).name == "Text Property"

    def test_can_parse_unquoted_property_names_with_spaces(self):
        node = " DOCPROPERTY Text Property  \\* MERGEFORMAT "
        assert FieldForTesting(node).name == "Text Property"

    def test_can_parse_quoted_property_names_with_extra_spaces(self):
        node = ' DOCPROPERTY  "Text Property"  \\* MERGEFORMAT '
        assert FieldForTesting(node).name == "Text Property"

    def test_can_parse_unquoted_property_names_with_extra_spaces(self):
        node = " DOCPROPERTY  Text Property  \\* MERGEFORMAT "
        assert FieldForTesting(node).name == "Text Property"


class TestFieldDateFormatParsing(object):
    def test_can_parse_quoted_date_format(self):
        node = ' DOCPROPERTY "Propertyname" \\@ "ddd-yy-MM" \\* MERGEFORMAT '
        assert FieldForTesting(node).name == "Propertyname"
        assert FieldForTesting(node).date_format == "E-yy-MM"

    def test_can_parse_quoted_date_format_with_spaces(self):
        node = ' DOCPROPERTY "Propertyname" \\@ "ddd yy MM" \\* MERGEFORMAT '
        assert FieldForTesting(node).name == "Propertyname"
        assert FieldForTesting(node).date_format == "E yy MM"

    def test_can_parse_quoted_date_format_with_extra_spaces(self):
        node = ' DOCPROPERTY "Propertyname"   \\@ "ddd yy MM"   \\* MERGEFORMAT'
        assert FieldForTesting(node).name == "Propertyname"
        assert FieldForTesting(node).date_format == "E yy MM"

    def test_can_parse_unquoted_date_format(self):
        node = ' DOCPROPERTY "Propertyname" \\@ DDD-yy-MM \\* MERGEFORMAT '
        assert FieldForTesting(node).name == "Propertyname"
        assert FieldForTesting(node).date_format == "E-yy-MM"

    def test_can_parse_unquoted_date_format_with_spaces(self):
        node = ' DOCPROPERTY "Propertyname" \\@ dddd yy MMMM \\* MERGEFORMAT '
        assert FieldForTesting(node).name == "Propertyname"
        assert FieldForTesting(node).date_format == "EEEE yy MMMM"

    def test_can_parse_unquoted_date_format_with_extra_spaces(self):
        node = ' DOCPROPERTY "Propertyname"   \\@ dd yy MM   \\* MERGEFORMAT'
        assert FieldForTesting(node).name == "Propertyname"
        assert FieldForTesting(node).date_format == "dd yy MM"


class TestFieldDateFormatMapping(object):
    def test_correctly_maps_simple_date(self):
        date = datetime(2020, 11, 19)

        assert word_to_python_date_format("yy/MM/dd") == "yy/MM/dd"
        assert word_to_python_date_format("YY/MM/DD") == "yy/MM/dd"
        assert (
            format_datetime(date, word_to_python_date_format("yy/MM/dd"), locale="en")
            == "20/11/19"
        )

        assert word_to_python_date_format("yyyy/MM/dd") == "yyyy/MM/dd"
        assert word_to_python_date_format("YYYY/MM/DD") == "yyyy/MM/dd"
        assert (
            format_datetime(date, word_to_python_date_format("YYYY/MM/DD"), locale="en")
            == "2020/11/19"
        )

    def test_correctly_maps_date_padding(self):
        date = datetime(2001, 2, 4)

        assert word_to_python_date_format("yy/MM/dd") == "yy/MM/dd"
        assert (
            format_datetime(date, word_to_python_date_format("yy/MM/dd"), locale="en")
            == "01/02/04"
        )

        assert word_to_python_date_format("yy/M/d") == "yy/M/d"
        assert (
            format_datetime(date, word_to_python_date_format("yy/M/d"), locale="en")
            == "01/2/4"
        )

    def test_correctly_maps_date_with_weekday_and_month_name(self):
        date = datetime(2020, 11, 19, 23, 59, 43)

        assert word_to_python_date_format("ddd dd MMM yyyy") == "E dd MMM yyyy"
        assert (
            format_datetime(
                date, word_to_python_date_format("ddd dd MMM yyyy"), locale="en"
            )
            == "Thu 19 Nov 2020"
        )

        assert word_to_python_date_format("dddd dd MMMM yyyy") == "EEEE dd MMMM yyyy"
        assert (
            format_datetime(
                date, word_to_python_date_format("dddd dd MMMM yyyy"), locale="en"
            )
            == "Thursday 19 November 2020"
        )

    def test_correctly_maps_date_with_time(self):
        date = datetime(2020, 11, 19, 1, 9, 8)

        assert (
            word_to_python_date_format("ddd DD MMM YYYY HH:mm:ss")
            == "E dd MMM yyyy HH:mm:ss"
        )
        assert (
            format_datetime(
                date,
                word_to_python_date_format("ddd DD MMM YYYY HH:mm:ss"),
                locale="en",
            )
            == "Thu 19 Nov 2020 01:09:08"
        )

        assert (
            word_to_python_date_format("ddd DD MMM YYYY H:m:s") == "E dd MMM yyyy H:m:s"
        )
        assert (
            format_datetime(
                date, word_to_python_date_format("ddd DD MMM YYYY H:m:s"), locale="en"
            )
            == "Thu 19 Nov 2020 1:9:8"
        )



from io import BytesIO

import pytest
from aiohttp import FormData
from docx import Document
from utils import ComparableDocument
from utils import docx_path
from utils import FixtureDocument

from docxcompose.server import create_app


@pytest.fixture
async def http_client(aiohttp_client):
    return await aiohttp_client(create_app())


async def test_post_with_documents_returns_composed_document(http_client):
    files = {
        "master": open(docx_path("master.docx"), "rb"),
        "table": open(docx_path("table.docx"), "rb"),
    }
    resp = await http_client.post("/", data=files)
    assert resp.status == 200
    composed_doc = ComparableDocument(Document(BytesIO(await resp.read())))
    composed_fixture = FixtureDocument("table.docx")
    assert composed_doc == composed_fixture


async def test_get_returns_405(http_client):
    resp = await http_client.get("/")
    assert resp.status == 405
    text = await resp.text()
    assert text == "405: Method Not Allowed"


async def test_post_without_multipart_returns_400(http_client):
    resp = await http_client.post("/")
    assert resp.status == 400
    text = await resp.text()
    assert text == "Multipart request required"


async def test_post_returns_400_if_compose_fails(http_client):
    data = FormData()
    data.add_field("master", "foo", content_type="plain/text")
    data.add_field("table", "bar", content_type="plain/text")
    resp = await http_client.post("/", data=data)
    assert resp.status == 400
    text = await resp.text()
    assert text == "No documents provided"


async def test_post_returns_500_if_compose_fails(http_client):
    files = {
        "master": BytesIO(b"FOO"),
        "table": BytesIO(b"bar"),
    }
    resp = await http_client.post("/", data=files)
    assert resp.status == 500
    text = await resp.text()
    assert text == "Failed composing documents"


async def test_post_with_url_parameters(http_client):
    files = {
        "master": open(docx_path("master.docx"), "rb"),
        "table": open(docx_path("table.docx"), "rb"),
    }
    resp = await http_client.post("/?preserve_styles=1", data=files)
    assert resp.status == 200
    composed_doc = ComparableDocument(Document(BytesIO(await resp.read())))
    composed_fixture = FixtureDocument("table.docx")
    assert composed_doc == composed_fixture


async def test_healtcheck_returns_200(http_client):
    resp = await http_client.get("/healthcheck")
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"



from utils import ComposedDocument
from utils import FixtureDocument


def test_footnote():
    doc = FixtureDocument("footnote.docx")
    composed = ComposedDocument("master.docx", "footnote.docx")

    assert composed == doc


def test_footnotes_with_hyperlinks():
    doc = FixtureDocument("footnotes_with_hyperlinks.docx")
    composed = ComposedDocument("master.docx", "footnotes_with_hyperlinks.docx")

    assert composed == doc



from lxml import etree

from docxcompose.utils import xml_elements_equal


def test_xml_elements_are_equal():
    xml1 = """
    <root>
        <a x="1">Foo</a>
        <b>Bar</b>
        <ignore_me>123</ignore_me>
    </root>
    """
    xml2 = """
    <root>
        <b>Bar</b>
        <a x="1">Foo</a>
        <ignore_me>999</ignore_me>
    </root>
    """
    e1 = etree.fromstring(xml1)
    e2 = etree.fromstring(xml2)
    assert xml_elements_equal(e1, e2, ignored_tags=["ignore_me"]) is True


def test_xml_elements_are_not_equal():
    xml1 = """
    <root>
        <a x="1">Foo</a>
        <b>Bar</b>
        <ignore_me>123</ignore_me>
    </root>
    """
    xml2 = """
    <root>
        <b>Bar</b>
        <a x="2">Foo</a>
        <ignore_me>999</ignore_me>
    </root>
    """
    e1 = etree.fromstring(xml1)
    e2 = etree.fromstring(xml2)
    assert xml_elements_equal(e1, e2, ignored_tags=["ignore_me"]) is False



from utils import ComposedDocument
from utils import FixtureDocument


def test_shapes():
    doc = FixtureDocument("embedded_visio.docx")
    composed = ComposedDocument("master.docx", "embedded_visio.docx")

    assert composed == doc



from utils import ComposedDocument
from utils import FixtureDocument


def test_table():
    doc = FixtureDocument("table.docx")
    composed = ComposedDocument("master.docx", "table.docx")

    assert composed == doc



from utils import ComposedDocument
from utils import FixtureDocument


def test_smartart():
    doc = FixtureDocument("smart_art.docx")
    composed = ComposedDocument("master.docx", "smart_art.docx")

    assert composed == doc



import pytest
from docx import Document
from utils import ComposedDocument
from utils import docx_path
from utils import FixtureDocument

from docxcompose.composer import Composer
from docxcompose.utils import xpath


def test_abstractnums_from_styles_are_not_duplicated(multiple_numberings):
    anums = xpath(
        multiple_numberings.doc.part.numbering_part.element,
        ".//w:abstractNum[.//w:pStyle]",
    )
    assert len(anums) == 2


def test_restart_first_numbering(multiple_numberings):
    paragraphs = xpath(multiple_numberings.doc.element.body, ".//w:p")
    assert len(xpath(paragraphs[9], ".//w:numId")) == 1


def test_do_not_restart_numbering_of_bullets(mixed_numberings):
    paragraphs = xpath(mixed_numberings.doc.element.body, ".//w:p")
    assert len(xpath(paragraphs[10], ".//w:numId")) == 0


def test_do_not_break_on_custom_styled_numbering(custom_styled_numbering):
    assert custom_styled_numbering.doc.element.xpath(".//w:numId/@w:val") == ["2"] * 4


def test_preserve_zero_numbering_references(numberings_with_zero_reference):
    numberings_with_zero_ref = xpath(
        numberings_with_zero_reference.doc.element.body, './/w:p//w:numId[@w:val="0"]'
    )
    assert len(numberings_with_zero_ref) == 2


def test_restarts_numbering_if_sequence_is_split_across_elements(
    numbering_with_paragraphs,
):
    numbering_ids = [
        each.val
        for each in xpath(numbering_with_paragraphs.doc.element.body, ".//w:numId")
    ]

    assert numbering_ids == [3, 3, 3, 3, 4, 4, 4]


def test_restart_numbering_manages_shared_style_names(static_reseed):
    doc = FixtureDocument("common_stylename_different_id.docx")
    composed = ComposedDocument(
        "common_stylename_different_id1.docx", "common_stylename_different_id2.docx"
    )
    assert composed == doc


def test_numberings(static_reseed):
    doc = FixtureDocument("numberings.docx")
    composed = ComposedDocument("numberings.docx", "numberings.docx")

    assert composed == doc


def test_restart_numberings():
    doc = FixtureDocument("numberings_restart.docx")
    composed = ComposedDocument("numberings_restart.docx", "numberings_restart.docx")

    assert composed == doc


def test_numberings_styles():
    doc = FixtureDocument("numberings_styles.docx")
    composed = ComposedDocument("numberings_styles.docx", "numberings_styles.docx")

    assert composed == doc


def test_numbering_reference_to_numbering_zero():
    doc = FixtureDocument("numbering_reference_to_numbering_zero.docx")
    composed = ComposedDocument(
        "numbering_reference_to_numbering_zero.docx",
        "numbering_reference_to_numbering_zero.docx",
    )

    assert composed == doc


def test_restarts_numbering_for_all_elements_of_same_sequence():
    doc = FixtureDocument("numbering_with_paragraphs_in_between.docx")
    composed = ComposedDocument(
        "numbering_with_paragraphs_in_between.docx",
        "numbering_with_paragraphs_in_between.docx",
    )

    assert composed == doc


def test_preserves_list_styles_when_restarting_numberings():
    doc = FixtureDocument("broken_listing.docx")
    composed = ComposedDocument("broken_listing_master.docx", "broken_listing.docx")

    assert composed == doc


def test_preserves_list_styles_when_restarting_many_numberings():
    doc = FixtureDocument("broken_listing_many.docx")
    composed = ComposedDocument(
        "broken_listing_master.docx", "broken_listing_many.docx"
    )

    assert composed == doc


def test_preserves_list_styles_when_restarting_nested_numberings():
    doc = FixtureDocument("broken_listing_nested.docx")
    composed = ComposedDocument(
        "broken_listing_master.docx", "broken_listing_nested.docx"
    )

    assert composed == doc


@pytest.fixture
def numberings_with_zero_reference():
    composer = Composer(
        Document(docx_path("numbering_reference_to_numbering_zero.docx"))
    )
    composer.append(Document(docx_path("numbering_reference_to_numbering_zero.docx")))
    return composer


@pytest.fixture
def numberings_in_styles():
    composer = Composer(Document(docx_path("master.docx")))
    composer.append(Document(docx_path("numberings_styles.docx")))
    return composer


@pytest.fixture
def multiple_numberings():
    composer = Composer(Document(docx_path("numberings_styles.docx")))
    composer.append(Document(docx_path("numberings_styles.docx")))
    return composer


@pytest.fixture
def mixed_numberings():
    composer = Composer(Document(docx_path("numberings_restart.docx")))
    composer.append(Document(docx_path("numberings_restart.docx")))
    return composer


@pytest.fixture
def numbering_with_paragraphs():
    composer = Composer(Document(docx_path("master.docx")))
    composer.append(Document(docx_path("numbering_with_paragraphs_in_between.docx")))
    return composer


@pytest.fixture
def custom_styled_numbering():
    composer = Composer(Document(docx_path("master.docx")))
    composer.append(Document(docx_path("custom_list_style.docx")))
    return composer



import pytest
from docx import Document
from utils import ComparableDocument
from utils import ComposedDocument
from utils import docx_path
from utils import FixtureDocument

from docxcompose.composer import Composer


def test_contains_predefined_styles_in_masters_language(merged_styles):
    style_ids = [s.style_id for s in merged_styles.doc.styles]
    assert "Heading1" in style_ids
    assert "Heading1" in style_ids
    assert "Strong" in style_ids
    assert "Quote" in style_ids


def test_does_not_contain_predefined_styles_in_appended_language(merged_styles):
    style_ids = [s.style_id for s in merged_styles.doc.styles]
    assert "berschrift1" not in style_ids
    assert "berschrift2" not in style_ids
    assert "Fett" not in style_ids
    assert "Zitat" not in style_ids


def test_contains_custom_styles_from_both_docs(merged_styles):
    style_ids = [s.style_id for s in merged_styles.doc.styles]
    assert "MyStyle1" in style_ids
    assert "MyStyle1Char" in style_ids
    assert "MeineFormatvorlage" in style_ids
    assert "MeineFormatvorlageZchn" in style_ids


def test_contains_linked_styles(merged_styles):
    style_ids = [s.style_id for s in merged_styles.doc.styles]
    assert "QuoteChar" in style_ids


def test_merged_styles_de():
    doc = FixtureDocument("styles_de.docx")
    composed = ComposedDocument("styles_de.docx", "styles_en.docx")

    assert composed == doc


def test_merged_styles_en():
    doc = FixtureDocument("styles_en.docx")
    composed = ComposedDocument("styles_en.docx", "styles_de.docx")

    assert composed == doc


def test_styles_are_not_switched_for_first_numbering_element():
    doc = FixtureDocument("switched_listing_style.docx")
    composed = ComposedDocument(
        "master_switched_listing_style.docx", "switched_listing_style.docx"
    )

    assert composed == doc


def test_continue_when_no_styles():
    """Expects not to throw a type error"""
    ComposedDocument("aatmay.docx", "aatmay.docx")


def test_preserve_styles_with_same_id():
    composer = Composer(
        Document(docx_path("styles_preserve1.docx")), preserve_styles=True
    )
    composer.append(Document(docx_path("styles_preserve2.docx")))
    style_ids = [s.style_id for s in composer.doc.styles]
    assert "MyCustomStyle" in style_ids
    assert "MyCustomStyle_1" in style_ids

    expected = FixtureDocument("styles_preserve.docx")
    composed = ComparableDocument(composer.doc)
    assert composed == expected


def test_ignore_styles_with_same_id():
    composer = Composer(Document(docx_path("styles_preserve1.docx")))
    composer.append(Document(docx_path("styles_preserve2.docx")))
    style_ids = [s.style_id for s in composer.doc.styles]
    assert "MyCustomStyle" in style_ids
    assert "MyCustomStyle_1" not in style_ids


def test_preserve_styles_does_not_duplicate_identical_styles():
    composer = Composer(
        Document(docx_path("styles_preserve1.docx")), preserve_styles=True
    )
    composer.append(Document(docx_path("styles_preserve2.docx")))
    composer.append(Document(docx_path("styles_preserve2.docx")))
    composer.append(Document(docx_path("styles_preserve1.docx")))
    assert [
        s.style_id
        for s in composer.doc.styles
        if s.style_id.startswith("MyCustomStyle")
    ] == ["MyCustomStyle", "MyCustomStyleZchn", "MyCustomStyle_1"]


def test_retain_formatting_from_default_styles():
    composer = Composer(
        Document(docx_path("styles_default1.docx")), preserve_styles=True
    )
    composer.append(Document(docx_path("styles_default2.docx")))
    composed = ComparableDocument(composer.doc)
    expected = FixtureDocument("styles_default.docx")
    assert composed == expected


@pytest.fixture
def merged_styles():
    composer = Composer(Document(docx_path("styles_en.docx")))
    composer.append(Document(docx_path("styles_de.docx")))
    return composer



import re

from docx.oxml.ns import nsmap
from docx.oxml.xmlchemy import BaseOxmlElement


NS = {
    "a": "http://schemas.openxmlformats.org/drawingml/2006/main",
    "asvg": "http://schemas.microsoft.com/office/drawing/2016/SVG/main",
    "cp": "http://schemas.openxmlformats.org/officeDocument/2006/custom-properties",
    "o": "urn:schemas-microsoft-com:office:office",
    "pic": "http://schemas.openxmlformats.org/drawingml/2006/picture",
    "r": "http://schemas.openxmlformats.org/officeDocument/2006/relationships",
    "v": "urn:schemas-microsoft-com:vml",
    "vt": "http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes",
    "w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
    "wp": "http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing",
    "dgm": "http://schemas.openxmlformats.org/drawingml/2006/diagram",
}
nsmap.update(NS)


def xpath(element, xpath_str):
    """Performs an XPath query on the given element and returns all matching
    elements.
    Works with lxml.etree._Element and with
    docx.oxml.xmlchemy.BaseOxmlElement elements.
    """
    if isinstance(element, BaseOxmlElement):
        return element.xpath(xpath_str)
    else:
        return element.xpath(xpath_str, namespaces=NS)


# Format specifications for docproperties can be found at
# https://support.microsoft.com/en-us/office/format-field-results-baa61f5a-5636-4f11-ab4f-6c36ae43508c?ui=en-US&rs=en-US&ad=US#ID0EAABAAA=Date-Time_format_switch_(\@)
date_format_map = (
    ("Y", "y"),  # Upper or lower case Y are equivalent
    ("D", "d"),
    ("dddd", "EEEE"),
    ("ddd", "E"),
    ("am/pm", "AM/PM"),  # Upper or lower case are equivalent
    ("AM/PM", "a"),
)


def word_to_python_date_format(format_str):
    for word_format, python_format in date_format_map:
        format_str = re.sub(word_format, python_format, format_str)
    return format_str


def increment_name(name):
    increment_part = name.split("_")[-1]
    try:
        increment = int(increment_part)
    except ValueError:
        return f"{name}_1"
    return f"{name.removesuffix(increment_part)}{increment + 1}"


def to_bool(value):
    return value.lower() in ["1", "yes", "true", "on", "ok"]


def xml_elements_equal(
    left,
    right,
    ignored_tags=None,
    compare_text=True,
    compare_tail=False,
    compare_attributes=True,
):
    return xml_element_signature(
        left,
        ignored_tags=ignored_tags,
        compare_text=compare_text,
        compare_tail=compare_tail,
        compare_attributes=compare_attributes,
    ) == xml_element_signature(
        right,
        ignored_tags=ignored_tags,
        compare_text=compare_text,
        compare_tail=compare_tail,
        compare_attributes=compare_attributes,
    )


def xml_element_signature(
    element,
    ignored_tags=None,
    compare_text=True,
    compare_tail=False,
    compare_attributes=True,
    is_root=True,
):
    """
    Creates a canonical, recursive representation of an element.

    Child elements are included as a sorted list of signatures,
    so their order is irrelevant.
    """
    tag = element.tag
    attrs = tuple(sorted(element.attrib.items())) if compare_attributes else ()
    text = normalize_text(element.text) if compare_text else None
    tail = normalize_text(element.tail) if compare_tail else None

    child_signatures = []
    for child in element:
        if ignored_tags and child.tag in ignored_tags:
            continue

        child_signatures.append(
            xml_element_signature(
                child,
                ignored_tags=ignored_tags,
                compare_text=compare_text,
                compare_tail=compare_tail,
                compare_attributes=compare_attributes,
                is_root=False,
            )
        )
    child_signatures.sort()

    if is_root:
        return (None, None, None, None, tuple(child_signatures))
    else:
        return (tag, attrs, text, tail, tuple(child_signatures))


def normalize_text(value):
    if value is None:
        return ""
    return value.strip()



from utils import ComposedDocument
from utils import FixtureDocument


def test_hyperlinks():
    doc = FixtureDocument("embedded_excel_chart.docx")
    composed = ComposedDocument("master.docx", "embedded_excel_chart.docx")

    assert composed == doc



import pytest
from docx import Document
from utils import ComposedDocument
from utils import docx_path
from utils import FixtureDocument

from docxcompose.composer import Composer
from docxcompose.utils import xpath


def test_header_and_footer_refs_in_paragraph_props_get_removed(header_footer):
    refs = xpath(
        header_footer.doc.element.body,
        ".//w:pPr/w:sectPr/w:headerReference|.//w:pPr/w:sectPr/w:footerReference",
    )
    assert len(refs) == 0


def test_master_header_and_footer_are_preserved_when_adding_sections():
    doc = FixtureDocument("master_header_footer_with_sections.docx")
    composed = ComposedDocument(
        "master_header_footer.docx", "header_footer_sections.docx"
    )

    assert composed == doc


def test_header_footer():
    doc = FixtureDocument("header_footer.docx")
    composed = ComposedDocument("header_footer.docx", "header_footer.docx")
    assert composed == doc


def test_header_footer_sections():
    doc = FixtureDocument("header_footer_sections.docx")
    composed = ComposedDocument(
        "header_footer_sections.docx", "header_footer_sections.docx"
    )

    assert composed == doc


@pytest.fixture
def header_footer():
    composer = Composer(Document(docx_path("master.docx")))
    composer.append(Document(docx_path("header_footer_sections.docx")))
    return composer



from utils import ComposedDocument
from utils import FixtureDocument


def test_hyperlinks():
    doc = FixtureDocument("hyperlinks.docx")
    composed = ComposedDocument("master.docx", "hyperlinks.docx")

    assert composed == doc



import os.path
import sys
from argparse import ArgumentParser

from docx import Document

from docxcompose.composer import Composer


def setup_parser():
    parser = ArgumentParser(description="compose multiple docx files into one file.")
    parser.add_argument(
        "master",
        help="path to master template that defines styles, headings and so on",
    )
    parser.add_argument(
        "files",
        nargs="+",
        help="path to one or more word-files to be appended to the master template",
        metavar="file",
    )
    parser.add_argument(
        "-o",
        "--output-document",
        dest="ouput_document",
        default="composed.docx",
        help="path to the output file",
        metavar="file",
    )
    parser.add_argument(
        "--preserve-styles",
        action="store_true",
        default=False,
    )
    return parser


def require_valid_file(parser, path):
    if not os.path.isfile(path):
        parser.error(message="file not found {}".format(path))


def parse_args(parser, args):
    parsed_args = parser.parse_args(args=args)

    require_valid_file(parser, parsed_args.master)
    for file_path in parsed_args.files:
        require_valid_file(parser, file_path)

    return parsed_args


def compose_files(parser, parsed_args):
    options = {
        "preserve_styles": parsed_args.preserve_styles,
    }
    composer = Composer(Document(parsed_args.master), **options)
    for slave_path in parsed_args.files:
        composer.append(Document(slave_path))

    composer.save(parsed_args.ouput_document)
    parser.exit(
        message="successfully composed file at {}\n".format(parsed_args.ouput_document)
    )


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = setup_parser()
    parsed_args = parse_args(parser, args)
    compose_files(parser, parsed_args)



from utils import ComposedDocument
from utils import FixtureDocument


def test_images():
    doc = FixtureDocument("images.docx")
    composed = ComposedDocument("master.docx", "images.docx")

    assert composed == doc


def test_embedded_and_external_image():
    doc = FixtureDocument("embedded_and_external_image.docx")
    composed = ComposedDocument("master.docx", "embedded_and_external_image.docx")

    assert composed == doc


def test_renumbering_of_non_visual_properties():
    expected = FixtureDocument("renumbering_nv_props.docx")
    composed = ComposedDocument("header_with_image.docx", "image.docx")

    assert composed == expected



import pytest
from utils import docx_path

from docxcompose import command


def test_command_creates_composed_docx_file_at_output_path(tmpdir):
    output_path = tmpdir.join("outfile.docx")
    assert not output_path.exists()

    arguments = [
        docx_path("master.docx"),
        docx_path("table.docx"),
        "--output-document",
        output_path.strpath,
    ]
    with pytest.raises(SystemExit) as exc_info:
        command.main(arguments)

    assert exc_info.value.code == 0
    assert output_path.exists()
    assert output_path.isfile()
    assert output_path.size() > 0



from utils import ComposedDocument
from utils import FixtureDocument


def test_section_types_are_correct():
    composed = ComposedDocument(
        "continous_section_break.docx", "continous_section_break.docx"
    )
    assert [s.start_type for s in composed.doc.sections] == [2, 0, 0]


def test_continuous_continuous_section_break():
    doc = FixtureDocument("continous_section_break.docx")
    composed = ComposedDocument(
        "continous_section_break.docx", "continous_section_break.docx"
    )

    assert composed == doc


def test_continuous_odd_section_break():
    doc = FixtureDocument("continous_odd_section_break.docx")
    composed = ComposedDocument(
        "continous_section_break.docx", "odd_section_break.docx"
    )

    assert composed == doc


def test_next_page_section_break():
    doc = FixtureDocument("next_page_section_break.docx")
    composed = ComposedDocument("next_page_section_break.docx", "legacy_empty.docx")

    assert composed == doc


def test_next_page_section_break_in_otherwise_empty_file_preseved():
    doc = FixtureDocument("next_page_section_break_empty.docx")
    composed = ComposedDocument("master.docx", "next_page_section_break_empty.docx")

    assert composed == doc
